//----------------------------------------------------------------------------------
// aviutl GPU TCYtB^
// Copyright (c) 2008  N099
//----------------------------------------------------------------------------------
#include <intrin.h>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <new>

#include <d3d9.h>
#include <d3dx9.h>
#include <d3dx9math.h>
#include <dxerr.h>

#include "resource.h"
#include "filter.h"

#include "prof.h"
#include "thrproc.h"
#include "weight_reader.h"
#include "gpufilter.h"

#include "weight_size.h"
#include "gpu_resize.fxh" //RpCς݃GtFNg

//---------------------------------------------------------------------
// const
//---------------------------------------------------------------------
//D3DXCui[hd3dxdll̃o[Wwj
const int D3DXLIB_VER_MIN = 38;
const int D3DXLIB_VER_MAX = 42;

//dݗpT[tFX
const D3DFORMAT FMT_WEIGHT = D3DFMT_V16U16;

//---------------------------------------------------------------------
// struct
//---------------------------------------------------------------------
struct PPVERT
{
    float x, y, z, rhw;
    float u, v, q, w;       // Texcoord

    const static D3DVERTEXELEMENT9 Decl[3];
};

const D3DVERTEXELEMENT9 PPVERT::Decl[3] =
{
	{ 0, 0,  D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITIONT, 0 },
	{ 0, 16, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,  0 },
	D3DDECL_END()
};

struct MTCP_PARAM
{
	int h, w;
	int sPitch, ycPitch;
	BYTE *pSurf;
	PIXEL_YC *pYC;
};

//---------------------------------------------------------------------
// internal function prototypes
//---------------------------------------------------------------------
static void set_window_weight(BYTE*, int, const std::vector<float>&, int);
static inline void yc_i16_mmx(BYTE*, const PIXEL_YC*, unsigned int);
static inline void i16_yc_mmx(PIXEL_YC*, const BYTE*, unsigned int);
static inline void yc_u16_mmx(BYTE*, const PIXEL_YC*, unsigned int);
static inline void u16_yc_mmx(PIXEL_YC*, const BYTE*, unsigned int);
static void __cdecl cp_yc_i16_mt(int, int, void*, void*);
static void __cdecl cp_i16_yc_mt(int, int, void*, void*);
static void __cdecl cp_yc_u16_mt(int, int, void*, void*);
static void __cdecl cp_u16_yc_mt(int, int, void*, void*);

static volatile long mt_cnt;

//---------------------------------------------------------------------
// public functions
//---------------------------------------------------------------------

//RXgN^
GpuFilter::GpuFilter() :
	m_pD3D(NULL), m_pd3dDevice(NULL),
	m_pTexWeight(NULL), m_pTexSrc(NULL), m_pTexRt(NULL), m_pSrfRt(NULL),
	m_pTexWeightTmp(NULL), m_pTexSysSrc(NULL), m_pTexSysDst(NULL),
	m_pVert(NULL), m_pVertDec(NULL), m_pEffect(NULL), m_pQuery(NULL),
	m_tproc(NULL), m_hD3dxdll(NULL), m_pCreateEffect(NULL)
{
	m_wSrc = m_hSrc = m_wItm = m_hItm = m_wDst = m_hDst = -1;
	m_pycStride = m_wMax = m_hMax = -1;

	m_pixFormat = U16;
	ZeroMemory(m_pixFormatAvailable, sizeof(m_pixFormatAvailable));

	ZeroMemory(m_hParam, sizeof(m_hParam));
	m_hTec = NULL;
	ZeroMemory(m_hTex, sizeof(m_hTex));

	m_pass[0] = m_pass[1] = -1;
	m_weights.clear();
}

//fXgN^
GpuFilter::~GpuFilter()
{
	Release();
	delete m_tproc;

	while(!m_weights.empty()){
		delete m_weights.back();
		m_weights.pop_back();
	}

	if(m_hD3dxdll){
		FreeLibrary(m_hD3dxdll);
	}
}

//IuWFNg̏
bool GpuFilter::Init(FILTER *fp, int dev)
{
	HRESULT hr;

	//Ŝ
	Release();

	//Load D3DXLib
	if(FAILED(LoadD3DXFunc())) return false;

	//init D3D    
	m_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if(NULL == m_pD3D) {
		ShowMessageBox(__FILE__, (DWORD)__LINE__, E_FAIL, TEXT("Direct3DCreate9()s܂"));
		return false;
	}

	//adapter selection
	UINT adapter = (dev < 0) ? D3DADAPTER_DEFAULT : dev;
	D3DDEVTYPE type = (dev < 0) ? D3DDEVTYPE_REF : D3DDEVTYPE_HAL;
	if(adapter >= m_pD3D->GetAdapterCount()){
		adapter = D3DADAPTER_DEFAULT;
	}

	//create device
	D3DPRESENT_PARAMETERS d3dpp; 

	ZeroMemory(&d3dpp, sizeof(d3dpp));
	d3dpp.Windowed = TRUE;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;

	hr = m_pD3D->CreateDevice(
		adapter, type, fp->hwnd,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING|D3DCREATE_NOWINDOWCHANGES|D3DCREATE_MULTITHREADED, 
		&d3dpp, &m_pd3dDevice );
	if(FAILED(hr)) {
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DfoCX̍쐬Ɏs܂"));
		Release();
		return false;
	}

	//foCX̃`FbN
	if(FAILED(CheckDeviceCap())){
		Release();
		return false;
	}

	// ɑ݂郊\[X쐬

	//Xbhp
	SAFE_DELETE(m_tproc);
	m_tproc = new(std::nothrow) ThrProc;
	if(m_tproc == NULL){
		hr = E_POINTER;
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("Xbh쐬Ɏs܂"));
		Release();
		return false;
	}
	if(m_tproc->Init(AsyncProc, this)){
		hr = E_FAIL;
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("XbhɎs܂"));
		Release();
		return false;
	}

	//d݂
	AppendExtWeight("<Lanczos>");
	//AppendExtWeight("<Tent>");
	//AppendExtWeight("<Rect>");
	AppendExtWeight("<Rect&Tent>");
	AppendExtWeight("<PWCubic>");

	//GtFNg̍쐬
	hr = CreateEffect();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GtFNg̍쐬Ɏs܂"));
		Release();
		return false;
	}

	//_錾̍쐬
	SAFE_RELEASE(m_pVertDec);
    hr = m_pd3dDevice->CreateVertexDeclaration(PPVERT::Decl, &m_pVertDec);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("_錾̍쐬Ɏs܂"));
		Release();
		return false;
	}

#ifdef PROFILE_GPU
	//fobOp
    hr = CreateQuery();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("NG̍쐬Ɏs܂"));
		Release();
		return false;
	}
#endif

	return true;
}

//e탊\[X̊m
bool GpuFilter::AllocResources(int w0, int h0, int w1, int h1, int w_max, int fmt)
{
	//̃`FbN
	if(!m_pEffect) return false;

	//sNZTCỸeXgiŏ32x32j
	if(w0<32 || w0>m_wMax || h0<32 || h0>m_hMax ||
	   w1<32 || w1>m_wMax || h1<32 || h1>m_hMax ||
	   w_max<32 || w_max<w0 || w_max<w1 )
	{
		return false;
	}

	//sNZTCYZbg
	m_wSrc = w0;
	m_hSrc = h0;
	m_wDst = w1;
	m_hDst = h1;
	m_pycStride = w_max;
	SetIntermediateSize();

	//sNZtH[}bg
	SetPixelFmt(static_cast<GpuFilter::PIXEL_FORMAT>(fmt));
	
	//ꉞJĂ
	ReleaseResources();

	//VXeɍ쐬郊\[X
	if(FAILED(CreateSysMemResources())){
		return false;
	}

	//GPUɍ쐬郊\[X
	if(FAILED(CreateDefaultResources())){
		return false;
	}

	//GtFNgɃ\[XZbg
	if(FAILED(SetEffectResources())){
		return false;
	}

	return true;
}

//D3D\[X̊J
bool GpuFilter::Release()
{
	//\[X̊J
	SAFE_RELEASE(m_pQuery);
	SAFE_RELEASE(m_pEffect);
	SAFE_RELEASE(m_pVertDec);
	ReleaseResources();

	//AddRefĂ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);

	return true;
}

//e탊\[X̊J
bool GpuFilter::ReleaseResources()
{	
	//GPU\[X̊J
	ReleaseDefaultResources();

	//VXe\[X̊J
	SAFE_RELEASE(m_pTexWeightTmp);
	SAFE_RELEASE(m_pTexSysSrc);
	SAFE_RELEASE(m_pTexSysDst);

	return true;
}

//tB^p[^̃Zbg
bool GpuFilter::SetParams(int r, int mode)
{
	HRESULT hr;
	int rd; //ۂ̃sNZa

	//̃`FbN
	if(!m_pEffect || !m_pTexWeightTmp) return false;

	_mm_empty();

	//d݃[h𐧌
	mode = max(0, min(mode, (int)m_weights.size()-1));

	//"r"
	r = min(m_weights[mode]->GetRadiusNum(), max(1, r));

	//dݔz擾
	std::vector<float> wtbl;
	rd = m_weights[mode]->GetWeight(r, wtbl);

	//d݃eNX`ɒlZbg
	D3DLOCKED_RECT lRect;
	hr = m_pTexWeightTmp->LockRect(0, &lRect, NULL, D3DLOCK_NOSYSLOCK);
	if(S_OK == hr){
		set_window_weight(static_cast<BYTE*>(lRect.pBits), lRect.Pitch, wtbl, rd);
		m_pTexWeightTmp->UnlockRect(0);
		hr = m_pd3dDevice->UpdateTexture(m_pTexWeightTmp, m_pTexWeight);
	}
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("d݂̃ZbgɎs܂"));
	}

	//g厞͈̔͂Zbg
	if(m_hParam[PARAM_RAD]){
		hr = m_pEffect->SetInt(m_hParam[PARAM_RAD], rd);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("͈̓p[^̃ZbgɎs܂"));
		}
	}
	
	//k͈̔͂Zbg
	if(m_hParam[PARAM_REP]){
		INT q[2] = { //255܂łɐ
			min(255, (rd*m_wSrc + (m_wDst+1)/2)/m_wDst),
			min(255, (rd*m_hSrc + (m_hDst+1)/2)/m_hDst)  };

		hr = m_pEffect->SetIntArray(m_hParam[PARAM_REP], q, ARRAYSIZE(q));
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, 
				TEXT("JԂp[^̃ZbgɎs܂"));
		}
	}

	return true;
}

//foCX̏ԂeXgAKvȂ烊Zbg
bool GpuFilter::TestDeviceState(bool *need_realloc)
{
	HRESULT hr;

	if(!m_pd3dDevice){
		return false;
	}

	// D3D9 device ̏ԂeXg
	hr = m_pd3dDevice->TestCooperativeLevel();
	switch(hr)
	{
		case D3D_OK:
			break;
		case D3DERR_DEVICELOST:
#ifdef PROFILE_GPU
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
				TEXT("foCXXg܂Bs܂B"));
#endif
			return false;
		case D3DERR_DEVICENOTRESET:
			hr = Reset();
			if(FAILED(hr)){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
					TEXT("foCXZbgɎs܂"));
				return false;
			}else{
				*need_realloc = true;
			}
			break;
		default :
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
				TEXT("TestCooperativeLevel()G[Ԃ܂"));
			return false;
	}

	return true;
}

//PIXEL_YCf[^VXeT[tFXɃRs[
bool GpuFilter::PutImage(const PIXEL_YC *data, EXEC_MT func_mt)
{
	HRESULT hr;

	getCount(TEXT("TestDev "));
#ifdef USE_DYNAMIC_TEXTURE
	hr = PutPixelDataDT(m_pTexSrc, data, func_mt);
#else
	hr = PutPixelData(m_pTexSysSrc, data, func_mt);
#endif
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("T[tFXփRs[ł܂ł"));
		return false;
	}

	return true;
}

//f[^̏s
bool GpuFilter::ProcImage(int empty)
{
	if(!m_pSrfRt || !m_pEffect || !m_pd3dDevice || m_pass[0] < 0){
		return false;
	}

	//_OJn
	if(SUCCEEDED(m_pd3dDevice->BeginScene()))
	{
#ifndef USE_DYNAMIC_TEXTURE
		//f[^̃Rs[
		m_pd3dDevice->UpdateTexture(m_pTexSysSrc, m_pTexSrc);
#endif
		m_pEffect->SetTexture(m_hTex[TEX_SOURCE], m_pTexSrc);

		// Set the vertex declaration
		m_pd3dDevice->SetVertexDeclaration(m_pVertDec);

		// Draw the quad
		UINT cPasses;
		m_pEffect->Begin(&cPasses, 0);

		// 1 or 2 passes
		if(m_pass[1] < 0){
			// 1st pass only
			SetRenderTarget(m_pSrfRt);

			m_pEffect->BeginPass(m_pass[0]);
			m_pd3dDevice->SetStreamSource(0, m_pVert, 0, sizeof(PPVERT));
			m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
			m_pEffect->EndPass();
		}else{
			// 1st pass
			SetRenderTarget(m_pTexRt);

			m_pEffect->BeginPass(m_pass[0]);
			m_pd3dDevice->SetStreamSource(0, m_pVert, 0, sizeof(PPVERT));
			m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
			m_pEffect->EndPass();

			// 2nd pass
			SetRenderTarget(m_pSrfRt);
			m_pEffect->SetTexture(m_hTex[TEX_SOURCE], m_pTexRt);

			m_pEffect->BeginPass(m_pass[1]);
			m_pd3dDevice->SetStreamSource(0, m_pVert, 4*sizeof(PPVERT), sizeof(PPVERT));
			m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
			m_pEffect->EndPass();
		}

		m_pEffect->End();

		// End scene
		m_pd3dDevice->EndScene();

		// empty the command buffer
		if(empty){
			DWORD n;
			m_pd3dDevice->ValidateDevice(&n);
		}
	}

	return true;
}

//f[^̏siʃXbhŔ񓯊j
bool GpuFilter::ProcImageAsync()
{
	if(!m_pSrfRt || !m_pd3dDevice || !m_tproc){
		return false;
	}
	
	return m_tproc->Start() ? false : true;
}

//PIXEL_YCf[^foCX]
bool GpuFilter::GetImage(PIXEL_YC *data, EXEC_MT func_mt)
{
	HRESULT hr;

	if(!data || !m_tproc){
		return false;
	}

	if(m_tproc->isRunning()){
		//҂
		m_tproc->Wait();
	}else{
		//VXeɓ]
		hr = GetRenderTargetData();
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("RTf[^̃Rs[Ɏs܂"));
			return false;
		}
	}

#ifdef PROFILE_GPU
	getCount(TEXT("GetRT   "));
#endif

	//VXeϊăRs[
	hr = GetPixelData(data, m_pTexSysDst, func_mt);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("T[tFXRs[ł܂ł"));
		return false;
	}

	return true;
}

//eNX`tH[}bg̕\
const TCHAR *GpuFilter::GetPixelFormatStr(void) const
{
	switch(m_pixFormat){
		case S16:
			return TEXT("S16");
		case U16:
			return TEXT("U16");
		default:
			return TEXT("Unknown");
	}
}

//d݃t@C̓ǂݍ
bool GpuFilter::AppendExtWeight(const char *file)
{
	std::vector<BaseWeightReader*>::const_iterator ct = m_weights.begin();
	for( ; ct!=m_weights.end(); ++ct){
		if(**ct == file){
			return false;
		}
	}

	_mm_empty();

	BaseWeightReader *tmp = NULL;
	if(BaseWeightReader::LoadWeight(file, tmp)){
		m_weights.push_back(tmp);
		return true;
	}

	return false;
}

std::vector<std::string> GpuFilter::GetExtWeightFileNames() const
{
	std::vector<std::string> name;
	std::vector<BaseWeightReader*>::const_iterator ct = m_weights.begin();

	for( ; ct!=m_weights.end(); ++ct){
		if((*ct)->GetFileName() != NULL){
			name.push_back((*ct)->GetFileName());
		}
	}
	return name;
}

//͈͂̍ől
int GpuFilter::GetMaxRadius(int n) const
{
	n = max(0, min(n, (int)m_weights.size()-1));
	return m_weights[n]->GetRadiusNum();

}

int GpuFilter::GetWeightNum() const
{
	return m_weights.size();
}

const char *GpuFilter::GetWeightName(int n) const
{
	n = max(0, min(n, (int)m_weights.size()-1));
	return m_weights[n]->GetName();
}

//GPUł̏I܂ő҂
#ifdef PROFILE_GPU
void GpuFilter::WaitForIdle()
{
	if(m_pQuery){
		m_pQuery->Issue(D3DISSUE_END);
		while(S_FALSE == m_pQuery->GetData(NULL, 0, D3DGETDATA_FLUSH))
			__asm pause ;
	}
}
#endif

//---------------------------------------------------------------------
// private functions
//---------------------------------------------------------------------

//D3DXCů֐[h
HRESULT GpuFilter::LoadD3DXFunc()
{
	//NAĂ
	m_pCreateEffect = NULL;

	if(m_hD3dxdll == 0){
		//~ōŏɌCu𗘗p
		TCHAR name[64];
		for(int i=D3DXLIB_VER_MAX; i>=D3DXLIB_VER_MIN; --i){
#ifdef _DEBUG
	#ifdef UNICODE
			swprintf_s(name, L"d3dx9d_%02d.dll", i);
	#else
			sprintf_s(name, "d3dx9d_%02d.dll", i);
	#endif
#else
	#ifdef UNICODE
			swprintf_s(name, L"d3dx9_%02d.dll", i);
	#else
			sprintf_s(name, "d3dx9_%02d.dll", i);
	#endif
#endif
			//Cu[h
			m_hD3dxdll = LoadLibrary(name);
			if(m_hD3dxdll){
				break;
			}
		}
	}

	//[hłꍇA֐Gg擾
	if(m_hD3dxdll == 0){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, E_FAIL,
			TEXT("D3DXCu[hł܂B")
			TEXT("ŐVDirectXGh[U[^C𓱓ĂB"));
		return E_FAIL;
	}
	m_pCreateEffect = reinterpret_cast<CREATEEFFECT>
		(GetProcAddress(m_hD3dxdll, "D3DXCreateEffect"));

	//擾ł`FbN
	if(m_pCreateEffect){
		return S_OK;
	}else{
		FreeLibrary(m_hD3dxdll);
		m_hD3dxdll = 0;
		m_pCreateEffect = NULL;

		ShowMessageBox(__FILE__, (DWORD)__LINE__, E_FAIL,
			TEXT("D3DX֐̎擾Ɏs܂B")
			TEXT("ŐVDirectXGh[U[^C𓱓Ă݂ĂB"));
		return E_FAIL;
	}
}

//foCX̃T|[g󋵂̃`FbN
HRESULT GpuFilter::CheckDeviceCap()
{
	HRESULT hr;

	//|C^̃`FbN
	if(!m_pd3dDevice || !m_pD3D) return D3DERR_INVALIDCALL;

	//D3DCaps̃`FbN
	D3DCAPS9 caps;
	ZeroMemory(&caps, sizeof(caps));
	hr = m_pd3dDevice->GetDeviceCaps(&caps);
	if(D3D_OK != hr) return hr;

	//eNX`֘ÃeXg
	if(caps.TextureCaps & D3DPTEXTURECAPS_SQUAREONLY){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("񐳕eNX`𗘗pł܂"));
		return D3DERR_NOTAVAILABLE;
	}
	if(!(caps.TextureCaps & D3DPTEXTURECAPS_POW2)){
		; //2̗ݏȊÕTCỸeNX`
	}else if(!(caps.TextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("2̗ݏTCYȊÕeNX`͗pł܂"));
		return D3DERR_NOTAVAILABLE;
	}

	//őTCY
	m_hMax = caps.MaxTextureHeight;
	m_wMax = caps.MaxTextureWidth;

	//sNZVF[_̃eXg
	if(caps.PixelShaderVersion < D3DPS_VERSION(3,0)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("PS3.0ȍ~T|[gGPUKvł"));
		return D3DERR_NOTAVAILABLE;
	}

    //tH[}bg̃`FbN
    D3DDISPLAYMODE DisplayMode;
    hr = m_pd3dDevice->GetDisplayMode( 0, &DisplayMode );
	if(D3D_OK != hr) return hr;

	bool found = false;
	if(SUCCEEDED(CheckRtTexFormat(
		caps.AdapterOrdinal, caps.DeviceType, DisplayMode.Format, D3DFMT_A16B16G16R16)))
	{
		found = true;
		m_pixFormatAvailable[U16] = true;
		SetPixelFmt(U16);
	}else{
#ifdef PROFILE_GPU
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A16B16G16R16 pł܂"));
#endif
		m_pixFormatAvailable[U16] = false;
	}

	if(SUCCEEDED(CheckRtTexFormat(
		caps.AdapterOrdinal, caps.DeviceType, DisplayMode.Format, D3DFMT_Q16W16V16U16)))
	{
		found = true;
		m_pixFormatAvailable[S16] = true;
		SetPixelFmt(S16);
	}else{
#ifdef PROFILE_GPU
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_Q16W16V16U16 pł܂"));
#endif
		m_pixFormatAvailable[S16] = false;
	}

	if(!found){   
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("p\ȃsNZtH[}bgЂƂ܂"));
		return D3DERR_NOTAVAILABLE;
	}

	//d݃eNX`p
	if(!SUCCEEDED(m_pD3D->CheckDeviceFormat(
		caps.AdapterOrdinal, caps.DeviceType, DisplayMode.Format, 0, D3DRTYPE_TEXTURE, D3DFMT_V16U16)))
	{
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_V16U16 pł܂"));
		return D3DERR_NOTAVAILABLE;
	}

	return hr;
}

//_̐錾
HRESULT GpuFilter::CreateVert()
{
	HRESULT hr;

    //foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pVert);

	const PPVERT Quad[8] =
	{
		//1st pass
		{        -0.5f,        -0.5f, 1.0f, 1.0f, 0.0f, 0.0f,        -0.5f,        -0.5f },
		{ m_wItm -0.5f,        -0.5f, 1.0f, 1.0f, 1.0f, 0.0f, m_wSrc -0.5f,        -0.5f },
		{        -0.5f, m_hItm -0.5f, 1.0f, 1.0f, 0.0f, 1.0f,        -0.5f, m_hSrc -0.5f },
		{ m_wItm -0.5f, m_hItm -0.5f, 1.0f, 1.0f, 1.0f, 1.0f, m_wSrc -0.5f, m_hSrc -0.5f },
		//2nd pass
		{        -0.5f,        -0.5f, 1.0f, 1.0f, 0.0f, 0.0f,        -0.5f,        -0.5f },
		{ m_wDst -0.5f,        -0.5f, 1.0f, 1.0f, 1.0f, 0.0f, m_wItm -0.5f,        -0.5f },
		{        -0.5f, m_hDst -0.5f, 1.0f, 1.0f, 0.0f, 1.0f,        -0.5f, m_hItm -0.5f },
		{ m_wDst -0.5f, m_hDst -0.5f, 1.0f, 1.0f, 1.0f, 1.0f, m_wItm -0.5f, m_hItm -0.5f }
	};

	hr = m_pd3dDevice->CreateVertexBuffer(sizeof(Quad),
		D3DUSAGE_WRITEONLY, 0, D3DPOOL_DEFAULT, &m_pVert, NULL );

	//obt@ɓ]
	LPVOID pVBData;
	if(SUCCEEDED(m_pVert->Lock(0, 0, &pVBData, D3DLOCK_NOSYSLOCK))){
		CopyMemory(pVBData, Quad, sizeof(Quad));
		m_pVert->Unlock();
	}

	return hr;
}

//GtFNgt@CǂݍŃGtFNg쐬
HRESULT GpuFilter::CreateEffect()
{
	const DWORD ShaderFlags = D3DXFX_NOT_CLONEABLE | D3DXFX_LARGEADDRESSAWARE;
	ID3DXBuffer *pCompErr = NULL; //RpC̃G[
	HRESULT hr;

	//foCXƊ֐̃`FbN
	if(!m_pd3dDevice || !m_pCreateEffect) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pEffect);

	//GtFNgt@C̓ǂݎ
	hr = m_pCreateEffect(m_pd3dDevice, BINARY_EFFECT, sizeof(BINARY_EFFECT),
		NULL, NULL, ShaderFlags, NULL, &m_pEffect, &pCompErr);
	if(D3D_OK != hr){
		if(pCompErr){
			//RpCG[\
			char *msg = (char*)pCompErr->GetBufferPointer();
#ifdef UNICODE
			if(msg){
				size_t count = pCompErr->GetBufferSize();
				WCHAR *str = new(std::nothrow) WCHAR[count];
				if(str){
					mbstowcs_s(NULL, str, count, msg, count - 1);
					str[count - 1] = L'\0';
					ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, str);
					delete[] str;
				}
			}
#else
			if(msg){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, msg);
			}
#endif
			pCompErr->Release();
		}
		
	}else{
		//p[^nh̎擾
		m_hParam[PARAM_RAD] = m_pEffect->GetParameterByName(NULL, "iRadius");
		m_hParam[PARAM_REP] = m_pEffect->GetParameterByName(NULL, "iRep");
		m_hParam[PARAM_SSZ] = m_pEffect->GetParameterByName(NULL, "fRszSrc");
		m_hParam[PARAM_MRT] = m_pEffect->GetParameterByName(NULL, "fMinRt");

		//eNjbN
		m_hTec = m_pEffect->GetTechniqueByName("ResizeGPU");

		//eNX`
		m_hTex[TEX_WEIGHT] = m_pEffect->GetParameterByName(NULL, "txWgt");
		m_hTex[TEX_SOURCE] = m_pEffect->GetParameterByName(NULL, "txSrc");
	}

	SAFE_RELEASE(pCompErr);

	return hr;
}

//VXeɕێ郊\[X쐬
HRESULT GpuFilter::CreateSysMemResources()
{
	HRESULT hr;
#ifndef USE_DYNAMIC_TEXTURE
	//VXẽeNX`i摜j
	hr = CreateTexture(&m_pTexSysSrc, 0, D3DPOOL_SYSTEMMEM, m_wSrc, m_hSrc);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("VXeeNX`íj̍쐬Ɏs܂"));
		return hr;
	}
#endif
	//VXẽeNX`iój
	hr = CreateTexture(&m_pTexSysDst, 0, D3DPOOL_SYSTEMMEM, m_wDst, m_hDst);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("VXeeNX`iój̍쐬Ɏs܂"));
		return hr;
	}

	//VXẽeNX`id݁j
	hr = CreateTexture(&m_pTexWeightTmp, 0, D3DPOOL_SYSTEMMEM, WIDTH_WEIGHT, HEIGHT_WEIGHT);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("VXeeNX`id݁j̍쐬Ɏs܂"));
		return hr;
	}

	return hr;
}

//GPUɑ݂郊\[X쐬FXgɂ͕AKv
HRESULT GpuFilter::CreateDefaultResources(void)
{
	HRESULT hr;

	//_
	hr = CreateVert();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("_̃ZbgɎs܂"));
		return hr;
	}

	//\[XeNX`
#ifdef USE_DYNAMIC_TEXTURE
	hr = CreateTexture(&m_pTexSrc, D3DUSAGE_DYNAMIC, D3DPOOL_DEFAULT, m_wSrc, m_hSrc);
#else
	hr = CreateTexture(&m_pTexSrc, 0, D3DPOOL_DEFAULT, m_wSrc, m_hSrc);
#endif
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("eNX`íj̍쐬Ɏs܂"));
		return hr;
	}

	//RTeNX`isvȏꍇ͍쐬Ȃj
	if(m_wSrc != m_wDst && m_hSrc != m_hDst){
		hr = CreateTexture(&m_pTexRt, D3DUSAGE_RENDERTARGET, D3DPOOL_DEFAULT, m_wItm, m_hItm);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
				TEXT("eNX`iԁj̍쐬Ɏs܂"));
			return hr;
		}
	}

	//d݃eNX`
	hr = CreateTexture(&m_pTexWeight, 0, D3DPOOL_DEFAULT, WIDTH_WEIGHT, HEIGHT_WEIGHT);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("eNX`id݁j̍쐬Ɏs܂"));
		return hr;
	}

	//RT
	hr = CreateSurface(&m_pSrfRt, D3DUSAGE_RENDERTARGET, D3DPOOL_DEFAULT, m_wDst, m_hDst);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("RTT[tFX̍쐬Ɏs܂"));
		return hr;
	}

#ifdef PROFILE_GPU
	//fobOp
    hr = CreateQuery();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("NG̍쐬Ɏs܂"));
		return hr;
	}
#endif
	return hr;
}

//\[XGtFNgɃZbg
HRESULT GpuFilter::SetEffectResources()
{
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice || !m_pEffect) return D3DERR_INVALIDCALL;

	//eNX`̃Zbg
	hr = m_pEffect->SetTexture(m_hTex[TEX_WEIGHT], m_pTexWeight);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("d݃eNX`̃ZbgɎs܂"));
		return hr;
	}

	hr = m_pEffect->SetTexture(m_hTex[TEX_SOURCE], m_pTexSrc);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("\[XeNX`̃ZbgɎs܂"));
		return hr;
	}

	_mm_empty();

	//̓sb`̃Zbg
	D3DXVECTOR4 stmp(1.0f/m_wSrc, 1.0f/m_hSrc, 0.0f, 0.0f);
	hr = m_pEffect->SetVector(m_hParam[PARAM_SSZ], &stmp);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("̓sb`̃ZbgɎs܂"));
		return hr;
	}

	//TCỸZbg
	D3DXVECTOR4 rtmp(m_wDst/(float)m_wSrc, m_hDst/(float)m_hSrc, 0.0f, 0.0f);
	hr = m_pEffect->SetVector(m_hParam[PARAM_MRT], &rtmp);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("TCỸZbgɎs܂"));
		return hr;
	}

	//eNjbN̑I
	hr = m_pEffect->SetTechnique(m_hTec);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("eNjbÑZbgɎs܂"));
		return hr;
	}
	SelectPasses();

	return hr;
}

//GPU\[XJ
void GpuFilter::ReleaseDefaultResources()
{
	SAFE_RELEASE(m_pVert);
	SAFE_RELEASE(m_pTexWeight);
	SAFE_RELEASE(m_pTexSrc);
	SAFE_RELEASE(m_pTexRt);
	SAFE_RELEASE(m_pSrfRt);
#ifdef PROFILE_GPU
	SAFE_RELEASE(m_pQuery);
#endif
}

//foCXXgɃZbgs
HRESULT GpuFilter::Reset(void)
{
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice || !m_pEffect) return D3DERR_INVALIDCALL;

	//GPU\[X̊J
	ReleaseDefaultResources();
	m_pEffect->OnLostDevice();

	//Zbg
	D3DPRESENT_PARAMETERS dpp;

	ZeroMemory(&dpp, sizeof(dpp));
	dpp.Windowed = TRUE;
	dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	dpp.BackBufferFormat = D3DFMT_UNKNOWN;

	hr = m_pd3dDevice->Reset(&dpp);
	if(FAILED(hr)){
		return hr;
	}

	hr = m_pEffect->OnResetDevice();
	if(FAILED(hr)){
		return hr;
	}

	//\[X̍Ċmۂ͂ł͍sȂ
	return hr;
}

//eNX`쐬
HRESULT GpuFilter::CreateTexture(IDirect3DTexture9 **ppTex, DWORD usage, D3DPOOL pool, UINT w, UINT h)
{
	//==HEIGHT_WEIGHTȂdݗpƂ݂Ȃ
	D3DFORMAT fmt = (h == HEIGHT_WEIGHT) ? FMT_WEIGHT : PicFmtToD3DFmt(m_pixFormat);

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;
	
	//Ŝ
	SAFE_RELEASE(*ppTex);

	//eNX`쐬
	return m_pd3dDevice->CreateTexture(w, h, 1, usage, fmt, pool, ppTex, NULL);
}

//T[tFX쐬
HRESULT GpuFilter::CreateSurface(IDirect3DSurface9 **ppSrf, DWORD usage, D3DPOOL pool, UINT w, UINT h)
{
	//==HEIGHT_WEIGHTȂdݗpƂ݂Ȃ
	D3DFORMAT fmt = (h == HEIGHT_WEIGHT) ? FMT_WEIGHT : PicFmtToD3DFmt(m_pixFormat);

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(*ppSrf);

	if(usage & D3DUSAGE_RENDERTARGET){
		//_[^[Qbg쐬
		return m_pd3dDevice->CreateRenderTarget(w, h, fmt, D3DMULTISAMPLE_NONE, 0, FALSE, ppSrf, NULL);
	}else{
		//ItXN[T[tFX쐬
		return m_pd3dDevice->CreateOffscreenPlainSurface(w, h, fmt, pool, ppSrf, NULL);
	}
}

//UpdateTexture𗘗pԐړ]
HRESULT GpuFilter::PutPixelData(IDirect3DTexture9 *sys, const PIXEL_YC *src, EXEC_MT func_mt)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!m_pd3dDevice || !sys) return D3DERR_INVALIDCALL;

	//VXẽT[tFXbN
	hr = sys->LockRect(0, &lRect, NULL, D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	getCount(TEXT("PutLock "));

	//f[^̃Rs[
	MULTI_THREAD_FUNC yc_int16_mt = (m_pixFormat == S16) ? cp_yc_i16_mt : cp_yc_u16_mt;
	MTCP_PARAM para = { m_hSrc, m_wSrc, lRect.Pitch, m_pycStride,
		static_cast<BYTE*>(lRect.pBits), const_cast<PIXEL_YC*>(src) };
	if(func_mt){
		InterlockedExchange(&mt_cnt, 0);
		func_mt(yc_int16_mt, &para, NULL);
	}else{
		yc_int16_mt(0, 0, &para, NULL);
	}

	getCount(TEXT("PutCopy "));
	
	//f[^̓]Proc()̍ŏɍs
	return sys->UnlockRect(0);
}

//DynamicTexture𗘗pړ]
HRESULT GpuFilter::PutPixelDataDT(IDirect3DTexture9 *dst, const PIXEL_YC *src, EXEC_MT func_mt)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!dst) return D3DERR_INVALIDCALL;

	//_Ci~bNeNX`̃T[tFXbN
	hr = dst->LockRect(0, &lRect, NULL, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	getCount(TEXT("PutLock "));

	//f[^̓]
	MULTI_THREAD_FUNC yc_int16_mt = (m_pixFormat == S16) ? cp_yc_i16_mt : cp_yc_u16_mt;
	MTCP_PARAM para = { m_hSrc, m_wSrc, lRect.Pitch, m_pycStride,
		static_cast<BYTE*>(lRect.pBits), const_cast<PIXEL_YC*>(src) };
	if(func_mt){
		InterlockedExchange(&mt_cnt, 0);
		func_mt(yc_int16_mt, &para, NULL);
	}else{
		yc_int16_mt(0, 0, &para, NULL);
	}

	getCount(TEXT("PutCopy "));

	return dst->UnlockRect(0);
}

//GetRenderTargetData𗘗pԐړ]
HRESULT GpuFilter::GetRenderTargetData()
{
	HRESULT hr;

	if(!m_pd3dDevice){
		return D3DERR_INVALIDCALL;
	}

	//Rs[擾
	IDirect3DSurface9 *dst = NULL;
	hr = m_pTexSysDst->GetSurfaceLevel(0, &dst);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VXeT[tFX擾ł܂ł"));
		return hr;
	}
#if 0
	//eNX`T[tFX̎擾
	IDirect3DSurface9 *src = NULL;
	hr = m_pTexRt->GetSurfaceLevel(0, &src);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("RTeNX`T[tFX擾ł܂ł"));
		SAFE_RELEASE(dst);
		return hr;
	}
	//VXeɓ]
	hr = m_pd3dDevice->GetRenderTargetData(src, dst);
	SAFE_RELEASE(src);
	SAFE_RELEASE(dst);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("foCXRs[ł܂ł"));
		return hr;
	}
#else
	//VXeɓ]
	hr = m_pd3dDevice->GetRenderTargetData(m_pSrfRt, dst);
	SAFE_RELEASE(dst);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("foCXRs[ł܂ł"));
		return hr;
	}
#endif

	return hr;
}

//VXẽf[^ϊăRs[
HRESULT GpuFilter::GetPixelData(PIXEL_YC *dst, IDirect3DTexture9 *sys, EXEC_MT func_mt)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!m_pd3dDevice || !sys) return D3DERR_INVALIDCALL;

	//VXẽT[tFXbN
	hr = sys->LockRect(0, &lRect, NULL, D3DLOCK_READONLY|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	getCount(TEXT("GetLock "));

	//f[^̃Rs[
	MULTI_THREAD_FUNC int16_yc_mt = (m_pixFormat == S16) ? cp_i16_yc_mt : cp_u16_yc_mt;
	MTCP_PARAM para = { m_hDst, m_wDst, lRect.Pitch, m_pycStride,
		static_cast<BYTE*>(lRect.pBits), dst };
	if(func_mt){
		InterlockedExchange(&mt_cnt, 0);
		func_mt(int16_yc_mt, &para, NULL);
	}else{
		int16_yc_mt(0, 0, &para, NULL);
	}

	getCount(TEXT("GetCopy "));

	return sys->UnlockRect(0);
}

//G[̕\
HRESULT GpuFilter::ShowMessageBox(const char *file, DWORD line, HRESULT hr, const TCHAR *msg)
{
#ifdef _DEBUG
	return DXTrace(file, line, hr, msg, TRUE);
#else
	MessageBox(NULL, msg, TEXT("GPU TCYtB^"), MB_ICONERROR|MB_SYSTEMMODAL);
	return S_OK;
#endif
}

//eNX`p
HRESULT GpuFilter::UpdateSurface(IDirect3DSurface9 *src, IDirect3DTexture9 *dst)
{
	HRESULT hr;
	IDirect3DSurface9 *srf = NULL;

	if(!src || !dst || !m_pd3dDevice){
		return D3DERR_INVALIDCALL;
	}

	//T[tFX̎擾
	hr = dst->GetSurfaceLevel(0, &srf);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("eNX`T[tFX擾ł܂ł"));
		return hr;
	}

	hr = m_pd3dDevice->UpdateSurface(src, NULL, srf, NULL);
	srf->Release();

	return hr;
}

//eNX`p
HRESULT GpuFilter::SetRenderTarget(IDirect3DTexture9 *dst)
{
	HRESULT hr;
	IDirect3DSurface9 *srf = NULL;

	if(!dst || !m_pd3dDevice){
		return D3DERR_INVALIDCALL;
	}

	//T[tFX̎擾
	hr = dst->GetSurfaceLevel(0, &srf);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("eNX`T[tFX擾ł܂ł"));
		return hr;
	}
	
	//RenderTarget̐ݒ
	hr = m_pd3dDevice->SetRenderTarget(0, srf);
	srf->Release();

	return hr;
}

//T[tFXp
HRESULT GpuFilter::SetRenderTarget(IDirect3DSurface9 *dst)
{
	if(!dst || !m_pd3dDevice){
		return D3DERR_INVALIDCALL;
	}else{
		return m_pd3dDevice->SetRenderTarget(0, dst);
	}
}

//sNZtH[}bg̎w
HRESULT GpuFilter::SetPixelFmt(PIXEL_FORMAT fmt)
{
	if(fmt < S16 || fmt >= Unknown) return E_FAIL;
	
	switch(fmt){
		case S16:
			if(m_pixFormatAvailable[fmt]){
				m_pixFormat = S16;
				break;
			}
		case U16:
		case Unknown:
		default:
			m_pixFormat = U16;
	}

	return S_OK;
}

//tH[}bgp\ǂ̃eXg
HRESULT GpuFilter::CheckRtTexFormat(UINT ad, D3DDEVTYPE dt, D3DFORMAT af, D3DFORMAT tgt) const
{
	if(!m_pD3D){
		return D3DERR_INVALIDCALL;
	}
	
	HRESULT hr;
	hr = m_pD3D->CheckDeviceFormat(ad, dt, af, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, tgt);
	if(S_OK != hr) return hr;
	hr = m_pD3D->CheckDeviceFormat(ad, dt, af, D3DUSAGE_RENDERTARGET, D3DRTYPE_SURFACE, tgt);
	if(S_OK != hr) return hr;
	hr = m_pD3D->CheckDeviceFormat(ad, dt, af, D3DUSAGE_DYNAMIC, D3DRTYPE_TEXTURE, tgt);
	if(S_OK != hr) return hr;
	hr = m_pD3D->CheckDeviceFormat(ad, dt, af, 0, D3DRTYPE_TEXTURE, tgt);

	return hr;
}

//2pX̒ԃTCY
void GpuFilter::SetIntermediateSize()
{
	if(m_wSrc == m_wDst || m_hSrc == m_hDst){
		//1passiŏITCYƓj
		m_wItm = m_wDst;
		m_hItm = m_hDst;
	}else{
		//2pass
		if(m_wSrc > m_wDst){
			//ɏkiDj
			m_wItm = m_wDst;
			m_hItm = m_hSrc;
		}else{
			//Ɋgi͌񂵁j
			m_wItm = m_wSrc;
			m_hItm = m_hDst;
		}
	}
}

//o͂ƒԃTCYpXݒ
void GpuFilter::SelectPasses()
{
	//Ȃ疳ȃoX
	m_pass[0] = m_pass[1] = -1;

	if(m_wSrc == m_wDst || m_hSrc == m_hDst){
		//1pass
		if(m_wSrc > m_wDst)		 m_pass[0] = MIN_X; //kX
		else if(m_wSrc < m_wDst) m_pass[0] = MAG_X; //gX
		else if(m_hSrc > m_hDst) m_pass[0] = MIN_Y; //kY
		else if(m_hSrc < m_hDst) m_pass[0] = MAG_Y; //gY
	}else{
		//2pass,
		if(m_wSrc > m_wDst){
			m_pass[0] = MIN_X; //1st:kX
			m_pass[1] = (m_hSrc > m_hDst) ? MIN_Y : MAG_Y;
		}else{
			m_pass[1] = MAG_X; //2nd:gX
			m_pass[0] = (m_hSrc > m_hDst) ? MIN_Y : MAG_Y;
		}
	}
}

//PIXEL_FORMATD3DFORMATɕϊ
D3DFORMAT GpuFilter::PicFmtToD3DFmt(PIXEL_FORMAT p) const
{
	switch(p){
		case S16:
			return D3DFMT_Q16W16V16U16;
		case U16:
			return D3DFMT_A16B16G16R16;
		case Unknown:
		default:
			return D3DFMT_UNKNOWN;
	}
}

//ʃXbhŏ
void __cdecl GpuFilter::AsyncProc(void *param)
{
	GpuFilter *pF = static_cast<GpuFilter*>(param);
	
	if(pF){
		if(pF->ProcImage()){
			pF->GetRenderTargetData();
		}
	}
}

#ifdef PROFILE_GPU
//fobOpNG
HRESULT GpuFilter::CreateQuery()
{
	SAFE_RELEASE(m_pQuery);
	return m_pd3dDevice->CreateQuery(D3DQUERYTYPE_EVENT, &m_pQuery);
}
#endif

//---------------------------------------------------------------------
// internal functions
//---------------------------------------------------------------------

inline SHORT f2short(float f)
{
	f = (f * 32767.0f) + 0.5f;
	f = min(32767.0f, max(-32767.0f, f));
	return static_cast<SHORT>(floor(f));
}

//d݂Zbgi[_␳j
static void 
set_window_weight(BYTE *b, int pitch, const std::vector<float> &w, int r)
{
	//[_̃eNZS܂łL͈͂ƍl
	int j = 0;
	for( ; j<r; ++j){
		SHORT *p = reinterpret_cast<SHORT*>(b);
		for(int i=0; i<WIDTH_WEIGHT; i++){
			short sh = f2short(w[j*WIDTH_WEIGHT + i]);
			p[2*(WIDTH_WEIGHT-i)-1] = p[2*i] = sh;
		}
		b += pitch;
	}
	for( ; j<HEIGHT_WEIGHT; ++j){
		SHORT *p = reinterpret_cast<SHORT*>(b);
		ZeroMemory(p, sizeof(SHORT) * 2 * WIDTH_WEIGHT);
		b += pitch;
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_Q16W16V16U16 MMXp
void yc_i16_mmx(BYTE *dst, const PIXEL_YC *src, unsigned int n)
{
	unsigned int i;
	__m64 *mdst = (__m64*)dst;

	for(i=0; i<n; ++i){
		*mdst++ = *(__m64*)src++;
		//_mm_stream_pi(mdst++, *(__m64*)src++);
	}
}

//ϊRs[ D3DFMT_Q16W16V16U16 -> PIXEL_YC MMXp
void i16_yc_mmx(PIXEL_YC *dst, const BYTE *src, unsigned int n)
{
	unsigned int i;
	const __m64 *msrc = (const __m64*)src;

	for(i=0; i<n; ++i){
		*(__m64*)dst++ = *msrc++;
	}
}

//ϊRs[(line) PIXEL_YC -> D3DFMT_A16B16G16R16 MMXp
void yc_u16_mmx(BYTE *dst, const PIXEL_YC *src, unsigned int n)
{
	unsigned int i;
	__m64 *mdst = (__m64*)dst;
	const __m64 u16_ofs = { 0x8000800080008000UI64 };

	for(i=0; i<n; ++i){
		*mdst++ = _mm_add_pi16(*(__m64*)src++, u16_ofs);
		//_mm_stream_pi(mdst++, _mm_add_pi16(*(__m64*)src++, u16_ofs));
	}
}

//ϊRs[ D3DFMT_A16B16G16R16 -> PIXEL_YC MMXp
void u16_yc_mmx(PIXEL_YC *dst, const BYTE *src, unsigned int n)
{
	unsigned int i;
	const __m64 *msrc = (const __m64*)src;
	const __m64 u16_ofs = { 0x8000800080008000UI64 };

	for(i=0; i<n; ++i){
		*(__m64*)dst++ = _mm_sub_pi16(*msrc++, u16_ofs);
	}
}

//}`Xbhp PIXEL_YC -> D3DFMT_Q16W16V16U16
static void __cdecl cp_yc_i16_mt(int, int num, void *p0, void*)
{
	const MTCP_PARAM *para = static_cast<MTCP_PARAM*>(p0);
	long i;
	BYTE *dst = para->pSurf;
	PIXEL_YC *src = para->pYC;

	if(num){
		while((i = _InterlockedExchangeAdd(&mt_cnt, 1)) < para->h){
			yc_i16_mmx(dst + i * para->sPitch, src + i * para->ycPitch, para->w);
		}
	}else{
		for(i=0; i<para->h; ++i){
			yc_i16_mmx(dst, src, para->w);
			dst += para->sPitch;
			src += para->ycPitch;
		}
	}

	_mm_empty();
}

//}`Xbhp D3DFMT_Q16W16V16U16 -> PIXEL_YC
static void __cdecl cp_i16_yc_mt(int, int num, void *p0, void*)
{
	const MTCP_PARAM *para = static_cast<MTCP_PARAM*>(p0);
	long i;
	BYTE *src = para->pSurf;
	PIXEL_YC *dst = para->pYC;

	if(num){
		while((i = _InterlockedExchangeAdd(&mt_cnt, 1)) < para->h){
			i16_yc_mmx(dst + i * para->ycPitch, src + i * para->sPitch, para->w);
		}
	}else{
		for(i=0; i<para->h; ++i){
			i16_yc_mmx(dst, src, para->w);
			dst += para->ycPitch;
			src += para->sPitch;
		}
	}

	_mm_empty();
}

//}`Xbhp D3DFMT_A16B16G16R16 -> PIXEL_YC
static void __cdecl cp_yc_u16_mt(int, int num, void *p0, void*)
{
	const MTCP_PARAM *para = static_cast<MTCP_PARAM*>(p0);
	long i;
	BYTE *dst = para->pSurf;
	PIXEL_YC *src = para->pYC;

	if(num){
		while((i = _InterlockedExchangeAdd(&mt_cnt, 1)) < para->h){
			yc_u16_mmx(dst + i * para->sPitch, src + i * para->ycPitch, para->w);
		}
	}else{
		for(i=0; i<para->h; ++i){
			yc_u16_mmx(dst, src, para->w);
			dst += para->sPitch;
			src += para->ycPitch;
		}
	}

	_mm_empty();
}

//}`Xbhp D3DFMT_A16B16G16R16 -> PIXEL_YC
static void __cdecl cp_u16_yc_mt(int, int num, void *p0, void*)
{
	const MTCP_PARAM *para = static_cast<MTCP_PARAM*>(p0);
	long i;
	BYTE *src = para->pSurf;
	PIXEL_YC *dst = para->pYC;

	if(num){
		while((i = _InterlockedExchangeAdd(&mt_cnt, 1)) < para->h){
			u16_yc_mmx(dst + i * para->ycPitch, src + i * para->sPitch, para->w);
		}
	}else{
		for(i=0; i<para->h; ++i){
			u16_yc_mmx(dst, src, para->w);
			dst += para->ycPitch;
			src += para->sPitch;
		}
	}

	_mm_empty();
}
